!* ========================================================================== *
!*
!* SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
!* http://software.intel.com/en-us/articles/
!*        intel-sample-source-code-license-agreement/
!*
!* Copyright 2014 Intel Corporation
!*
!* THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED,
!* INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY,
!* FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL
!* PROPERTY RIGHTS.
!*
!* ========================================================================== *
#ifdef __MIC__
!DIR$ message "You are compiling for the coprocessor now!"
#endif

! 1D linear convolution - the actual work
subroutine convolve(input, kernel, output, inputSize, kernelSize)
    implicit none
    integer i, j
    integer inputSize, kernelSize
    real*4 input(inputSize + kernelSize - 1), kernel(kernelSize), &
&          output(inputSize)
    real*4 res

!$OMP parallel do
    do i = 1, inputSize
        res = 0.0
        do j = 1, kernelSize
            res = res + kernel(j) * input(i + j - 1)
        enddo
        output(i) = res
    enddo
end subroutine

program native
    use omp_lib
    implicit none
    integer i
    integer argc
    integer inputSize, kernelSize
    integer numThreads
    character(len=256), allocatable :: argv(:)
    integer starttime, stoptime, rate
    real*4, allocatable :: input(:), kernel(:), output(:)
    logical inputOK, kernelOK, outputOK

    argc = command_argument_count()
    argc = argc + 1
    allocate(argv(argc))
    do i = 1, argc
        call get_command_argument(i - 1, argv(i))
    end do

    if (argc .ne. 4) then
        print *, "Error: Wrong use!"
        print *, "Usage: ", argv(1), " <input> <kernel> <output>"
        deallocate(argv)
        stop 1
    endif

    ! Get sizes of input and kernel and open input, kernel and output file
    inquire(file=argv(2), OPENED=inputOK, SIZE=inputSize)
    inputSize = inputSize / 4;
    if (inputOK) then
        print *, "Error: Cannot read input file!"
        deallocate(argv)
        stop 1
    endif
    if (inputSize .eq. 0) then
        print *, "Error: Input file is empty!"
        deallocate(argv)
        stop 1
    endif
    print *, "# elements of input file: ", inputSize
    open(file=argv(2), unit=8, access="stream",form="unformatted")

    inquire(file=argv(3), OPENED=kernelOK, SIZE=kernelSize)
    kernelSize = kernelSize / 4;
    if (kernelOK) then
        print *, "Error: Cannot read kernel file!"
        deallocate(argv)
        close(8)
        stop 1
    endif
    if (kernelSize .eq. 0) then
        print *, "Error: Kernel file is empty!"
        deallocate(argv)
        close(8)
        stop 1
    endif
    print *, "# elements of kernel file: ", kernelSize
    if (inputSize .lt. kernelSize) then
        print *, "Error: Input is smaller than kernel!"
        deallocate(argv)
        close(8)
        stop 1
    endif
    open(file=argv(3), unit=9, access="stream",form="unformatted")

    inquire(file=argv(4), OPENED=outputOK)
    if (outputOK) then
        print *, "Error: Cannot read kernel file!"
        deallocate(argv)
        close(8)
        close(9)
        stop 1
    endif
    open(file=argv(4), unit=10, access="stream",form="unformatted")
    deallocate(argv)

    allocate(input(inputSize + kernelSize - 1))
    allocate(kernel(kernelSize))
    allocate(output(inputSize))

    ! Read in input and kernel
    do i = 1, kernelSize - 1
        input(i) = 0.0 ! Fade in
    enddo
    read(8) input(kernelSize:inputSize + kernelSize - 1)
    close(8)
    read(9) kernel
    close(9)

    ! Initialize the OpenMP runtime once and inform about available threads
    call kmp_set_defaults("KMP_AFFINITY=scatter");
!$OMP parallel do
    do i = 1, 10
        if (i .eq. 1) then
            numThreads = OMP_GET_NUM_THREADS()
        endif
    enddo
    print *, "Number of OpenMP threads used: ", numThreads

    call system_clock(starttime, rate)
!DIR$ noinline ! Make sure compiler does not optimize over call
    call convolve(input, kernel, output, inputSize, kernelSize)
    call system_clock(stoptime, rate)

    write(10) output
    close(10)

    print *, "Elapsed time = ", (stoptime - starttime) / real(rate), " seconds"

    deallocate(input)
    deallocate(kernel)
    deallocate(output)
end program
